#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/mman.h>
#include "globals.h"
#include "synchronize.h"

// Buffer mutex
Mutex buffer_mutex;

// Semaphores
Semaphore empty_slots_available;
Semaphore filled_slots_available;

void produce(int id, int count);

// Critical region.
void insert_item(Item item);

// Pointer to the shared memory object.
Shared *shared;

int main(int argc, char *argv[])
{
    if (argc != 3)
    {
        fprintf(stderr, "Usage: id count\n");
        return -1;
    }

    int id        = atoi(argv[1]);
    int max_items = atoi(argv[2]);

    printf("                                      ID: %d\n", id);
    printf("                               Max items: %d\n", max_items);

    int shm_fd;  // shared memory file descriptor

    // If this is Producer #1, remove any
    // shared memory object and semaphores
    // with the same names.
    if (id == 1)
    {
        shm_unlink(SHARED_MEMORY_NAME);
        REMOVE(EMPTY_SEMAPHORE_NAME);
        REMOVE(FILLED_SEMAPHORE_NAME);

        // Create new semaphores.
        printf(" Empty slots available semaphore created: %p\n",
               (char *)CREATE(&empty_slots_available, EMPTY_SEMAPHORE_NAME,
                                                      BUFFER_SIZE));
        printf("Filled slots available semaphore created: %p\n",
               (char *)CREATE(&filled_slots_available, FILLED_SEMAPHORE_NAME,
                                                       0));

        // Initialize the buffer mutex lock.
        printf("                       Buffer mutex init: %p\n",
               (char *)INIT(&buffer_mutex, "buffer_mutex"));
    }
    else
    {
        // Open existing semaphores.
        printf("  Empty slots available semaphore opened: %p\n",
               (char *)OPEN(&empty_slots_available, EMPTY_SEMAPHORE_NAME));
        printf(" Filled slots available semaphore opened: %p\n",
               (char *)OPEN(&filled_slots_available, FILLED_SEMAPHORE_NAME));

        // Use the existing buffer mutex lock.
        printf("                       Buffer mutex used: %p\n",
               (char *)USE(&buffer_mutex, "buffer_mutex"));
    }

    // Create the shared memory object,
    // configure its size, and map it.
    shm_fd = shm_open(SHARED_MEMORY_NAME, O_CREAT | O_RDWR, 0666);
    ftruncate(shm_fd, SHARED_SIZE);
    shared = mmap(0, SHARED_SIZE, PROT_WRITE, MAP_SHARED, shm_fd, 0);

    printf("\n");

    produce(id, max_items);

    printf("\nDone!\n");
    return 0;
}

void produce(int id, int max_items)
{
    srand(time(0));
    Item item = 0;

    for (int i = 1; i <= max_items; i++)
    {
        sleep(rand()%2);
        item = 100*id + i;

        insert_item(item);

        printf("Producer %d inserted %5d\n", id, item);
        fflush(stdout);
    }
}

void insert_item(Item item)
{
    WAIT(&empty_slots_available);

    // Critical region: Insert an item into the buffer.
    LOCK(&buffer_mutex);
        shared->buffer[shared->in] = item;
        shared->in = (shared->in + 1)%BUFFER_SIZE;
    UNLOCK(&buffer_mutex);

    SIGNAL(&filled_slots_available);
}
